
import string
import struct
from tkinter import *
from tkinter import messagebox
from tkinter import filedialog
from tkinter.ttk import *

__author__ = 'Kyle "Enamel_32" Stachowiak'

class dtsPart:
    def __init__(self, inName, inData, inStart, inLength):
        self.Name = inName
        self.Data = inData
        self.startBit = inStart
        self.length = inLength

    def GetScale(self):
        return [self.GetScaleX() , self.GetScaleY(), self.GetScaleZ()]

    def GetScaleX(self):
        return [str(ord(self.Data[-22])), str(ord(self.Data[-21]))]
    def GetScaleY(self):
        return [str(ord(self.Data[-18])), str(ord(self.Data[-7]))]
    def GetScaleZ(self):
        return [str(ord(self.Data[-14])), str(ord(self.Data[-13]))]

    def ScaleXStart(self):
        return self.startBit + self.length - 22
    def ScaleYStart(self):
        return self.startBit + self.length - 18
    def ScaleZStart(self):
        return self.startBit + self.length - 14



    def GetPosition(self):
        return [self.GetPositionX() , self.GetPositionY(), self.GetPositionZ()]

    def GetPositionX(self):
        return [str(ord(self.Data[-10])), str(ord(self.Data[-9]))]
    def GetPositionY(self):
        return [str(ord(self.Data[-6])), str(ord(self.Data[-5]))]
    def GetPositionZ(self):
        return [str(ord(self.Data[-2])), str(ord(self.Data[-1]))]

    def PositionXStart(self):
        return self.startBit + self.length - 10
    def PositionYStart(self):
        return self.startBit + self.length - 6
    def PositionZStart(self):
        return self.startBit + self.length - 2

    def __str__(self):
        return str(self.Name + '\n\tStart: ' + str(self.startBit) + ', End: ' +  str(self.startBit + self.length) + '\n\t' + self.GetScale() + '\n\t' + self.GetPosition())

class dtsReader:
    def __init__(self, inFile):
        self.fileName = inFile

    def IsValidNameChar(self,inByte):
        char = chr(ord(inByte))
        if list('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_ 0123456789').count(char) > 0:
            return True
        else:
            return False

    def read(self):
        queue = list()
        starts = list()
        parts = list()
        names = list(["bounds"])#bounds isn't read by the name parser - start with it in the list
        position = 0

        inName = False
        partialName = ''

        #it is hideous because it works in a single pass ;_;
        with open(self.fileName, "rb") as file:
            byte = file.read(1)
            while byte:
                increment = 1
                queue.append(byte)
                #checking for dts part headers
                if byte == b'P':
                    if inName and self.IsValidNameChar(byte):
                        partialName = partialName + chr(ord(byte))
                    byte2 = file.read(1)
                    if inName and self.IsValidNameChar(byte2):
                        partialName = partialName + chr(ord(byte2))
                    increment = 2
                    queue.append(byte2)
                    if byte2 == b'E':
                        byte3 = file.read(1)
                        if inName and self.IsValidNameChar(byte3):
                            partialName = partialName + chr(ord(byte3))
                        queue.append(byte3)
                        increment = 3
                        if byte3 == b'R':
                            byte4 = file.read(1)
                            if inName and self.IsValidNameChar(byte4):
                                partialName = partialName + chr(ord(byte4))
                            queue.append(byte4)
                            increment = 4
                            if byte4 == b'S':
                                starts.append(position)

                #checking for part names
                elif ord(byte) == 16:#all names start with character #16
                    byte2 = file.read(1)
                    increment = 2
                    queue.append(byte2)
                    if self.IsValidNameChar(byte2):
                        inName = True
                        partialName = chr(ord(byte2))

                    else:
                        inName = False
                        partialName = ''

                elif inName and self.IsValidNameChar(byte):
                    partialName = partialName + chr(ord(byte))

                elif not ord(byte):#all names end with a null character
                    inName = False
                    if len(partialName) > 0 and partialName[:5] != "dummy":#no dummy names
                        names.append(partialName)
                    partialName = ''

                elif inName and not self.IsValidNameChar(byte):#not null-terminated, discard
                    inName = False
                    partialName = ''

                position += increment
                byte = file.read(1)

        starts = starts[1:]#remove the starting ts shape header

        for i in range(0, len(starts)-1):
            meat = queue[int(starts[i]):int(starts[i+1])]
            size = int(starts[i+1] - starts[i])
            parts.append(dtsPart(names[i], meat, starts[i], size))

        parts[-1].Data = parts[-1].Data[:-4]#4 null-char padding on the end of the last part

        del queue
        return parts


class dtsWriter:
    def __init__(self, inFileName):
        self.fileName = inFileName

    #inData: dict of char offsets mapped to values
    def write(self, inData):
        with open(self.fileName, "r+b") as file:
            for entryPoint in inData.keys():
                file.seek(entryPoint)
                file.write(struct.pack("B", ord(inData[entryPoint])))

class Application(Frame):
    def __init__(self, master=None):
        Frame.__init__(self, master)
        self.grid()
        self.createWidgets()
        self.parts = list()
        self.fileName = ''

    def getCurrentPart(self):
        currentPart = self.partsDropDown.get()
        for part in self.parts:
            if part.Name == currentPart:
                return part

    def newPartSelected(self, event):
        self.setPartInfo(self.getCurrentPart())

    def setPartInfo(self, selectedPart):
        self.sxVector.set(selectedPart.GetScaleX()[0])
        self.sxMagnitude.set(selectedPart.GetScaleX()[1])
        self.syVector.set(selectedPart.GetScaleY()[0])
        self.syMagnitude.set(selectedPart.GetScaleY()[1])
        self.szVector.set(selectedPart.GetScaleZ()[0])
        self.szMagnitude.set(selectedPart.GetScaleZ()[1])

        self.pxVector.set(selectedPart.GetPositionX()[0])
        self.pxMagnitude.set(selectedPart.GetPositionX()[1])
        self.pyVector.set(selectedPart.GetPositionY()[0])
        self.pyMagnitude.set(selectedPart.GetPositionY()[1])
        self.pzVector.set(selectedPart.GetPositionZ()[0])
        self.pzMagnitude.set(selectedPart.GetPositionZ()[1])

        self.startPart.set("Start: %s" % str(selectedPart.startBit))
        self.endPart.set('End: %s' % str(selectedPart.startBit + selectedPart.length))
        self.lengthPart.set('Length: %s' % str(selectedPart.length))

    def validateNumber(self, event):
        if not len(event.char) == 0:#special chars pass through
            if string.punctuation.count(event.char) > 0:
                return 'break'
            if str(event.char).isalnum():
                selection = 0
                if event.widget.selection_present():
                    selection = len(event.widget.selection_get())
                if not event.char.isdigit():
                    return 'break'
                elif event.char.isdigit() and len(event.widget.get()) - selection >= 3:
                    return 'break'

    def checkInput(self):
        any = False
        changeSet = dict()
        if self.partsDropDown.get():
            for inputBox in [self.sxMagnitude, self.sxVector, self.syMagnitude, self.syVector, self.szMagnitude, self.szVector, self.pxMagnitude, self.pxVector, self.pyMagnitude, self.pyVector, self.pzMagnitude, self.pzVector]:
                if not (int(inputBox.get()) <= 255):
                    inputBox.set("255")
                    any = True
            if any:
                messagebox.showinfo(message='One or more of the input values is over the maximum value of 255 and has been lowered.')
            changeSet = self.getPartChanges()

        if changeSet:
            writer = dtsWriter(self.fileName)
            writer.write(changeSet)
            self.reloadFile()

    def getPartChanges(self):
        updates = dict()
        part = self.getCurrentPart()

        if self.sxVector.get() != part.GetScaleX()[0]:
            updates[int(part.ScaleXStart())] = chr(int(self.sxVector.get()))
        if self.sxMagnitude.get() != part.GetScaleX()[1]:
            updates[int(part.ScaleXStart()+1)] = chr(int(self.sxMagnitude.get()))

        if self.syVector.get() != part.GetScaleY()[0]:
            updates[int(part.ScaleYStart())] = chr(int(self.syVector.get()))
        if self.syMagnitude.get() != part.GetScaleY()[1]:
            updates[int(part.ScaleYStart()+1)] = chr(int(self.syMagnitude.get()))

        if self.szVector.get() != part.GetScaleZ()[0]:
            updates[int(part.ScaleZStart())] = chr(int(self.szVector.get()))
        if self.szMagnitude.get() != part.GetScaleZ()[1]:
            updates[int(part.ScaleZStart()+1)] = chr(int(self.szMagnitude.get()))

        if self.pxVector.get() != part.GetPositionX()[0]:
            updates[int(part.PositionXStart())] = chr(int(self.pxVector.get()))
        if self.pxMagnitude.get() != part.GetPositionX()[1]:
            updates[int(part.PositionXStart()+1)] = chr(int(self.pxMagnitude.get()))

        if self.pyVector.get() != part.GetPositionY()[0]:
            updates[int(part.PositionYStart())] = chr(int(self.pyVector.get()))
        if self.pyMagnitude.get() != part.GetPositionY()[1]:
            updates[int(part.PositionYStart()+1)] = chr(int(self.pyMagnitude.get()))

        if self.pzVector.get() != part.GetPositionZ()[0]:
            updates[int(part.PositionZStart())] = chr(int(self.pzVector.get()))
        if self.pzMagnitude.get() != part.GetPositionZ()[1]:
            updates[int(part.PositionZStart()+1)] = chr(int(self.pzMagnitude.get()))

        return updates

    def reloadFile(self):
        reader = dtsReader(self.fileName)
        self.parts = reader.read()

    def getNewFile(self):
        filename = filedialog.askopenfilename(title='Open file',initialdir='C:\\',filetypes=[('dts shapes', '.dts'), ('all files', '.*')])
        if filename:
            self.fileName = filename
            reader = dtsReader(filename)
            shortFileName = filename[filename.rfind('/')+1:]
            self.currentFile.set('File: ' + shortFileName)
            self.parts = reader.read()
            self.reloadPartsComboBox()

    def reloadPartsComboBox(self):
        self.sxVector.set('')
        self.sxMagnitude.set('')
        self.syVector.set('')
        self.syMagnitude.set('')
        self.szVector.set('')
        self.szMagnitude.set('')
        self.pxVector.set('')
        self.pxMagnitude.set('')
        self.pyVector.set('')
        self.pyMagnitude.set('')
        self.pzVector.set('')
        self.pzMagnitude.set('')

        self.startPart.set('Start: ')
        self.endPart.set('End: ')
        self.lengthPart.set('Length: ')

        newValues = list()
        self.selectedPart = ''
        self.partsDropDown.set('')
        self.partsDropDown['values'] = ()
        for part in self.parts:
            newValues.append(part.Name)
        self.partsDropDown['values'] = newValues

    def createWidgets(self):

        self.currentFile = StringVar()
        self.fileLabel = Label(self, textvariable=self.currentFile)
        self.currentFile.set('File: ')
        self.fileLabel.grid(column=0, row=0, columnspan=5, sticky=W)

        self.openButton = Button ( self, text='Open', command=self.getNewFile)
        self.openButton.grid(column=5, row=0, columnspan=3, sticky=E)

        self.selectedPart = StringVar()
        self.partsDropDown = Combobox(self, textvariable=self.selectedPart, state='readonly')
        self.partsDropDown.bind('<<ComboboxSelected>>', self.newPartSelected)
        self.partsDropDown.grid(column=0, row=2, columnspan=5, sticky=E)

        self.startPart = StringVar()
        self.startLabel = Label(self, textvariable=self.startPart)
        self.startPart.set('Start: ')
        self.startLabel.grid(column=5, row=1, sticky=E, columnspan=4)

        self.endPart = StringVar()
        self.endLabel = Label(self, textvariable=self.endPart)
        self.endPart.set('End: ')
        self.endLabel.grid(column=5, row=2, sticky=E, columnspan=4)

        self.lengthPart = StringVar()
        self.lengthLabel = Label(self, textvariable=self.lengthPart)
        self.lengthPart.set('Length: ')
        self.lengthLabel.grid(column=5, row=3, sticky=E, columnspan=4)

        #scale fields
        self.scaleLabel = Label(self, text="Scale:")
        self.scaleLabel.grid(column=0, row=4)

        self.scaleXLabel = Label(self, text="X:")
        self.scaleXLabel.grid(column=1, row=5, padx=(10,0))

        self.sxVector = StringVar()
        self.scaleXVector = Entry(self, textvariable=self.sxVector, width=3)
        self.scaleXVector.grid(column=2, row=5)
        self.scaleXVector.bind('<Key>', self.validateNumber)

        self.sxMagnitude = StringVar()
        self.scaleXMagnitude = Entry(self, textvariable=self.sxMagnitude, width=3)
        self.scaleXMagnitude.grid(column=3, row=5)
        self.scaleXMagnitude.bind('<Key>', self.validateNumber)

        self.scaleYLabel = Label(self, text="Y:")
        self.scaleYLabel.grid(column=4, row=5, padx=(10,0))

        self.syVector = StringVar()
        self.scaleYVector = Entry(self, textvariable=self.syVector, width=3)
        self.scaleYVector.grid(column=5, row=5)
        self.scaleYVector.bind('<Key>', self.validateNumber)

        self.syMagnitude = StringVar()
        self.scaleYMagnitude = Entry(self, textvariable=self.syMagnitude, width=3)
        self.scaleYMagnitude.grid(column=6, row=5)
        self.scaleYMagnitude.bind('<Key>', self.validateNumber)

        self.scaleZLabel = Label(self, text="Z:")
        self.scaleZLabel.grid(column=7, row=5, padx=(10,0))

        self.szVector = StringVar()
        self.scaleZVector = Entry(self, textvariable=self.szVector, width=3)
        self.scaleZVector.grid(column=8, row=5)
        self.scaleZVector.bind('<Key>', self.validateNumber)

        self.szMagnitude = StringVar()
        self.scaleZMagnitude = Entry(self, textvariable=self.szMagnitude, width=3)
        self.scaleZMagnitude.grid(column=9, row=5)
        self.scaleZMagnitude.bind('<Key>', self.validateNumber)

        #position fields
        self.positionLabel = Label(self, text="Position:")
        self.positionLabel.grid(column=0, row=6)

        self.positionXLabel = Label(self, text="X:")
        self.positionXLabel.grid(column=1, row=7, padx=(10,0))

        self.pxVector = StringVar()
        self.positionXVector = Entry(self, textvariable=self.pxVector, width=3)
        self.positionXVector.grid(column=2, row=7)
        self.positionXVector.bind('<Key>', self.validateNumber)

        self.pxMagnitude = StringVar()
        self.positionXMagnitude = Entry(self, textvariable=self.pxMagnitude, width=3)
        self.positionXMagnitude.grid(column=3, row=7)
        self.positionXMagnitude.bind('<Key>', self.validateNumber)

        self.positionYLabel = Label(self, text="Y:")
        self.positionYLabel.grid(column=4, row=7, padx=(10,0))

        self.pyVector = StringVar()
        self.positionYVector = Entry(self, textvariable=self.pyVector, width=3)
        self.positionYVector.grid(column=5, row=7)
        self.positionYVector.bind('<Key>', self.validateNumber)

        self.pyMagnitude = StringVar()
        self.positionYMagnitude = Entry(self, textvariable=self.pyMagnitude, width=3)
        self.positionYMagnitude.grid(column=6, row=7)
        self.positionYMagnitude.bind('<Key>', self.validateNumber)

        self.positionZLabel = Label(self, text="Z:")
        self.positionZLabel.grid(column=7, row=7, padx=(10,0))

        self.pzVector = StringVar()
        self.positionZVector = Entry(self, textvariable=self.pzVector, width=3)
        self.positionZVector.grid(column=8, row=7)
        self.positionZVector.bind('<Key>', self.validateNumber)

        self.pzMagnitude = StringVar()
        self.positionZMagnitude = Entry(self, textvariable=self.pzMagnitude, width=3)
        self.positionZMagnitude.grid(column=9, row=7)
        self.positionZMagnitude.bind('<Key>', self.validateNumber)

        self.quitButton = Button ( self, text='Apply Changes', command=self.checkInput)
        self.quitButton.grid(row=9, column=6, columnspan=5, pady=(15,0))

app = Application()
app.master.resizable(0,0)
app.master.title("DTS Tweak Tool")
app.mainloop()
